"""The driver module provides :class:`Driver`\s for identifying and
reading data files.

This allows Hooke to be data-file agnostic.  Drivers for various
commercial force spectroscopy microscopes are provided, and it's easy
to write your own to handle your lab's specific format.
"""

from ..config import Setting
from ..util.pluggable import IsSubclass, construct_graph


DRIVER_MODULES = [
#    ('csvdriver', True),
#    ('hdf5', True),
#    ('hemingclamp', True),
#    ('jpk', True),
#    ('mcs', True),
#    ('mfp1dexport', True),
#    ('mfp3d', True),
#    ('picoforce', True),
#    ('picoforcealt', True),
    ('tutorial', True),
]
"""List of driver modules and whether they should be included by
default.  TODO: autodiscovery
"""

DRIVER_SETTING_SECTION = 'drivers'
"""Name of the config section which controls driver selection.
"""


class Driver(object):
    """Base class for file format drivers.
    
    :attr:`name` identifies your driver, and should match the module
    name.
    """
    def __init__(self, name):
        self.name = name
        self.setting_section = '%s driver' % self.name

    def dependencies(self):
        """Return a list of :class:`Driver`\s we require."""
        return []

    def default_settings(self):
        """Return a list of :class:`hooke.config.Setting`\s for any
        configurable driver settings.

        The suggested section setting is::

            Setting(section=self.setting_section, help=self.__doc__)
        """
        return []

    def is_me(self, path):
        """Read the file and return True if the filetype can be
        managed by the driver.  Otherwise return False.
        """
        return False

    def read(self, path):
        """Read data from `path` and return a
        (:class:`hooke.curve.Data`, `info`) tuple.

        The `info` :class:`dict` must contain values for the keys:
        'filetype' and 'experiment'.  See :class:`hooke.curve.Curve`
        for details.
        """
        raise NotImplementedError

# Construct driver dependency graph and load default drivers.

DRIVER_GRAPH = construct_graph(
    this_modname=__name__,
    submodnames=[name for name,include in DRIVER_MODULES],
    class_selector=IsSubclass(Driver, blacklist=[Driver]))
"""Topologically sorted list of all possible :class:`Driver`\s.
"""

def default_settings():
    settings = [Setting(DRIVER_SETTING_SECTION,
                        help='Enable/disable default drivers.')]
    for dnode in DRIVER_GRAPH:
        driver = dnode.data
        default_include = [di for mod_name,di in DRIVER_MODULES
                           if mod_name == driver.name][0]
        help = driver.__doc__.split('\n', 1)[0]
        settings.append(Setting(
                section=DRIVER_SETTING_SECTION,
                option=driver.name,
                value=str(default_include),
                help=help,
                ))
    for dnode in DRIVER_GRAPH:
        driver = dnode.data
        settings.extend(driver.default_settings())
    return settings
