#!/usr/bin/env python

'''
Hooke - A force spectroscopy review & analysis tool.

A free, open source data analysis platform

COPYRIGHT
'''

import multiprocessing
import optparse
import os.path
import unittest
import sys

from . import engine as engine
from . import config as config_mod
from . import playlist as playlist
from . import plugin as plugin_mod
from . import driver as driver_mod
from . import ui as ui


class Hooke (object):
    def __init__(self, config=None, debug=0):
        self.debug = debug
        default_settings = (config_mod.DEFAULT_SETTINGS
                            + plugin_mod.default_settings()
                            + driver_mod.default_settings()
                            + ui.default_settings())
        if config == None:
            config = config_mod.HookeConfigParser(
                paths=config_mod.DEFAULT_PATHS,
                default_settings=default_settings)
            config.read()
        self.config = config
        self.load_plugins()
        self.load_drivers()
        self.load_ui()
        self.command = engine.CommandEngine()

        self.playlists = playlist.NoteIndexList()

    def load_plugins(self):
        self.plugins = plugin_mod.load_graph(
            plugin_mod.PLUGIN_GRAPH, self.config, include_section='plugins')
        self.commands = []
        for plugin in self.plugins:
            self.commands.extend(plugin.commands())

    def load_drivers(self):
        self.drivers = plugin_mod.load_graph(
            driver_mod.DRIVER_GRAPH, self.config, include_section='drivers')

    def load_ui(self):
        self.ui = ui.load_ui(self.config)

    def close(self):
        self.config.write() # Does not preserve original comments

    def run(self):
        """Run Hooke's main execution loop.

        Spawns a :class:`hooke.engine.CommandEngine` subprocess and
        then runs the UI, rejoining the `CommandEngine` process after
        the UI exits.
        """
        ui_to_command = multiprocessing.Queue()
        command_to_ui = multiprocessing.Queue()
        command = multiprocessing.Process(
            target=self.command.run, args=(self, ui_to_command, command_to_ui))
        command.start()
        try:
            self.ui.run(self.commands, ui_to_command, command_to_ui)
        finally:
            ui_to_command.put(ui.CloseEngine())
            command.join()

    def run_lines(self, lines):
        """Run the pre-set commands `lines` with the "command line" UI.

        Allows for non-interactive sessions that are otherwise
        equivalent to :meth:'.run'.
        """
        cmdline = ui.load_ui(self.config, 'command line')
        ui_to_command = multiprocessing.Queue()
        command_to_ui = multiprocessing.Queue()
        command = multiprocessing.Process(
            target=self.command.run, args=(self, ui_to_command, command_to_ui))
        command.start()
        try:
            cmdline.run_lines(
                self.commands, ui_to_command, command_to_ui, lines)
        finally:
            ui_to_command.put(ui.CloseEngine())
            command.join()

def main():
    p = optparse.OptionParser()
    p.add_option(
        '-s', '--script', dest='script', metavar='FILE',
        help='Script of command line Hooke commands to run.')
    p.add_option(
        '-c', '--command', dest='commands', metavar='COMMAND',
        action='append', default=[],
        help='Add a command line Hooke command to run.')
    options,arguments = p.parse_args()
    if len(arguments) > 0:
        print >> sys.stderr, 'Too many arguments to %s: %d > 0' \
            % (sys.argv[0], len(arguments))
        print >> sys.stderr, p.help()
        sys.exit(1)

    app = Hooke(debug=__debug__)

    if options.script != None:
        f = open(os.path.expanduser(options.script), 'r')
        options.commands.extend(f.readlines())
        f.close
    if len(options.commands) > 0:
        app.run_lines(options.commands)
        sys.exit(0)

    try:
        app.run()
    finally:
        app.close()

if __name__ == '__main__':
    main()
