"""The `playlist` module provides :class:`PlaylistPlugin` several
associated :class:`hooke.command.Command`\s for handling
:mod:`hooke.playlist` classes.
"""

import glob

from ..command import Command, Argument, Failure
from ..playlist import FilePlaylist
from ..plugin import Builtin


class PlaylistPlugin (Builtin):
    def __init__(self):
        super(PlaylistPlugin, self).__init__(name='playlist')

    def commands(self):
        return [NextCommand(), PreviousCommand(), JumpCommand(),
                IndexCommand(), CurveListCommand(),
                SaveCommand(), LoadCommand(),
                AddCommand(), AddGlobCommand(),
                RemoveCommand(), FilterCommand(), NoteFilterCommand()]


# Define common or complicated arguments

def current_playlist_callback(hooke, command, argument, value):
    if value != None:
        return value
    playlist = hooke.playlists.current()
    if playlist == None:
        raise Failure('No playlists loaded')
    return playlist

PlaylistArgument = Argument(
    name='playlist', type='playlist', callback=current_playlist_callback,
    help="""
:class:`hooke.plugin.playlist.Playlist` to act on.  Defaults to the
current playlist.
""".strip())

def playlist_name_callback(hooke, command, argument, value):
        return hooke.playlists.free_name()

PlaylistNameArgument = Argument(
    name='name', type='string', optional=True, callback=playlist_name_callback,
    help="""
Name of the new playlist (defaults to an auto-generated name).
""".strip())

def all_drivers_callback(hooke, command, argument, value):
    return hooke.drivers


# Define commands

class NextCommand (Command):
    """Move playlist to the next curve.
    """
    def __init__(self):
        super(NextCommand, self).__init__(
            name='next curve',
            arguments=[PlaylistArgument],
            help=self.__doc__)

    def _run(self, hooke, inqueue, outqueue, params):
	params['playlist'].next()

class PreviousCommand (Command):
    """Move playlist to the previous curve.
    """
    def __init__(self):
        super(PreviousCommand, self).__init__(
            name='previous curve',
            arguments=[PlaylistArgument],
            help=self.__doc__)

    def _run(self, hooke, inqueue, outqueue, params):
	params['playlist'].previous()

class JumpCommand (Command):
    """Move playlist to a given curve.
    """
    def __init__(self):
        super(JumpCommand, self).__init__(
            name='jump to curve',
            arguments=[
                PlaylistArgument,
                Argument(name='index', type='int', optional=False, help="""
Index of target curve.
""".strip()),
                ],
            help=self.__doc__)

    def _run(self, hooke, inqueue, outqueue, params):
	params['playlist'].jump(params['index'])

class IndexCommand (Command):
    """Print the index of the current curve.

    The first curve has index 0.
    """
    def __init__(self):
        super(IndexCommand, self).__init__(
            name='curve index',
            arguments=[
                PlaylistArgument,
                ],
            help=self.__doc__)

    def _run(self, hooke, inqueue, outqueue, params):
	outqueue.put(params['playlist']._index)

class CurveListCommand (Command):
    """Get the curves in a playlist.
    """
    def __init__(self):
        super(CurveListCommand, self).__init__(
            name='playlist curves',
            arguments=[
                PlaylistArgument,
                ],
            help=self.__doc__)

    def _run(self, hooke, inqueue, outqueue, params):
	outqueue.put([c for c in params['playlist']])

class SaveCommand (Command):
    """Save a playlist.
    """
    def __init__(self):
        super(SaveCommand, self).__init__(
            name='save playlist',
            arguments=[
                PlaylistArgument,
                Argument(name='output', type='file',
                         help="""
File name for the output playlist.  Defaults to overwriting the input
playlist.
""".strip()),
                ],
            help=self.__doc__)

    def _run(self, hooke, inqueue, outqueue, params):
	params['playlist'].save(params['output'])

class LoadCommand (Command):
    """Load a playlist.
    """
    def __init__(self):
        super(LoadCommand, self).__init__(
            name='load playlist',
            arguments=[
                Argument(name='input', type='file', optional=False,
                         help="""
File name for the input playlist.
""".strip()),
                Argument(name='drivers', type='driver', optional=True,
                         count=-1, callback=all_drivers_callback,
                         help="""
Drivers for loading curves.
""".strip()),
                ],
            help=self.__doc__)

    def _run(self, hooke, inqueue, outqueue, params):
        p = FilePlaylist(drivers=params['drivers'], path=params['input'])
        p.load()
        hooke.playlists.append(p)
	outqueue.put(p)

class AddCommand (Command):
    """Add a curve to a playlist.
    """
    def __init__(self):
        super(AddCommand, self).__init__(
            name='add curve to playlist',
            arguments=[
                PlaylistArgument,
                Argument(name='input', type='file', optional=False,
                         help="""
File name for the input :class:`hooke.curve.Curve`.
""".strip()),
                Argument(name='info', type='dict', optional=True,
                         help="""
Additional information for the input :class:`hooke.curve.Curve`.
""".strip()),
                ],
            help=self.__doc__)

    def _run(self, hooke, inqueue, outqueue, params):
        params['playlist'].append_curve_by_path(params['input'],
                                                params['info'])

class AddGlobCommand (Command):
    """Add curves to a playlist with file globbing.

    Adding lots of files one at a time can be tedious.  With this
    command you can use globs (`data/curves/*.dat`) to add curves
    for all matching files at once.
    """
    def __init__(self):
        super(AddGlobCommand, self).__init__(
            name='glob curves to playlist',
            arguments=[
                PlaylistArgument,
                Argument(name='input', type='glob', optional=False,
                         help="""
File name glob for the input :class:`hooke.curve.Curve`.
""".strip()),
                Argument(name='info', type='dict', optional=True,
                         help="""
Additional information for the input :class:`hooke.curve.Curve`.
""".strip()),
                ],
            help=self.__doc__)

    def _run(self, hooke, inqueue, outqueue, params):
        for path in sorted(glob.glob(params['input'])):
            params['playlist'].append_curve_by_path(path, params['info'])

class RemoveCommand (Command):
    """Remove a curve from a playlist.
    """
    def __init__(self):
        super(RemoveCommand, self).__init__(
            name='remove curve from playlist',
            arguments=[
                PlaylistArgument,
                Argument(name='index', type='int', optional=False, help="""
Index of target curve.
""".strip()),
                ],
            help=self.__doc__)

    def _run(self, hooke, inqueue, outqueue, params):
        params['playlist'].pop(params['index'])
        params['playlist'].jump(params._index)

class FilterCommand (Command):
    """Create a subset playlist via a selection function.

    Removing lots of curves one at a time can be tedious.  With this
    command you can use a function `filter` to select the curves you
    wish to keep.
    """
    def __init__(self):
        super(FilterCommand, self).__init__(
            name='filter playlist',
            arguments=[
                PlaylistArgument,
                PlaylistNameArgument,
                Argument(name='filter', type='function', optional=False,
                         help="""
Function returning `True` for "good" curves.  `filter(curve) -> True/False`.
""".strip()),
                ],
            help=self.__doc__)

    def _run(self, hooke, inqueue, outqueue, params):
        p = params['playlist'].filter(params['filter'])
        hooke.playlists.add(p)
        outqueue.put(p)

class NoteFilterCommand (FilterCommand):
    """Create a subset playlist of curves with `.info['note'] != None`.
    """
    def __init__(self):
        super(NoteFilterCommand, self).__init__()
        self.name = 'note filter playlist'
        self.arguments = [a for a in self.arguments if a.name != 'filter']

    def _run(self, hooke, inqueue, outqueue, params):
        params['filter'] = lambda curve : \
            'note' in curve.info and curve.info['note'] != None
        return super(NoteFilterCommand, self)._run(
            hooke, inqueue, outqueue, params)
