"""The `ui` module provides :class:`UserInterface` and various subclasses.
"""

import ConfigParser as configparser

from .. import version
from ..compat.odict import odict
from ..config import Setting
from ..util.pluggable import IsSubclass


USER_INTERFACE_MODULES = [
    'commandline',
    #'gui',
    ]
"""List of user interface modules.  TODO: autodiscovery
"""

USER_INTERFACE_SETTING_SECTION = 'user interfaces'
"""Name of the config section which controls UI selection.
"""


class QueueMessage (object):
    def __str__(self):
        return self.__class__.__name__

class CloseEngine (QueueMessage):
    pass

class CommandMessage (QueueMessage):
    """A message storing a command to run, `command` should be a
    :class:`hooke.command.Command` instance, and `arguments` should be
    a :class:`dict` with `argname` keys and `value` values to be
    passed to the command.
    """
    def __init__(self, command, arguments):
        self.command = command
        self.arguments = arguments

class UserInterface (object):
    """A user interface to drive the :class:`hooke.engine.CommandEngine`.
    """
    def __init__(self, name):
        self.name = name
        self.setting_section = '%s user interface' % (self.name)
        self.config = {}

    def default_settings(self):
        """Return a list of :class:`hooke.config.Setting`\s for any
        configurable UI settings.

        The suggested section setting is::

            Setting(section=self.setting_section, help=self.__doc__)
        """
        return []

    def reload_config(self, config):
        """Update the user interface for new config settings.

        Should be called with the new `config` upon recipt of
        `ReloadUserInterfaceConfig` from the `CommandEngine`.
        """
        try:
            self.config = dict(config.items(self.setting_section))
        except configparser.NoSectionError:
            self.config = {}

    def run(self, hooke, ui_to_command_queue, command_to_ui_queue):
        return

    # Assorted useful tidbits for subclasses

    def _splash_text(self):
        return ("""
Hooke version %s

COPYRIGHT
----
""" % version()).strip()

    def _playlist_status(self, playlist):
        if len(playlist) > 0:
            return '%s (%s/%s)' % (playlist.name, playlist._index + 1,
                                   len(playlist))
        return 'The playlist %s does not contain any valid force curve data.' \
            % self.name

def construct_odict(this_modname, submodnames, class_selector):
    """Search the submodules `submodnames` of a module `this_modname`
    for class objects for which `class_selector(class)` returns
    `True`.  These classes are instantiated and stored in the returned
    :class:`hooke.compat.odict.odict` in the order in which they were
    discovered.
    """
    instances = odict()
    for submodname in submodnames:
        count = len([s for s in submodnames if s == submodname])
        assert count > 0, 'No %s entries: %s' % (submodname, submodnames)
        assert count == 1, 'Multiple (%d) %s entries: %s' \
            % (count, submodname, submodnames)
        this_mod = __import__(this_modname, fromlist=[submodname])
        submod = getattr(this_mod, submodname)
        for objname in dir(submod):
            obj = getattr(submod, objname)
            if class_selector(obj):
                instance = obj()
                instances[instance.name] = instance
    return instances

USER_INTERFACES = construct_odict(
    this_modname=__name__,
    submodnames=USER_INTERFACE_MODULES,
    class_selector=IsSubclass(UserInterface, blacklist=[UserInterface]))
""":class:`hooke.compat.odict.odict` of :class:`UserInterface`
instances keyed by `.name`.
"""

def default_settings():
    settings = [Setting(USER_INTERFACE_SETTING_SECTION,
                        help='Select the user interface (only one).')]
    for i,ui in enumerate(USER_INTERFACES.values()):
        help = ui.__doc__.split('\n', 1)[0]
        settings.append(Setting(USER_INTERFACE_SETTING_SECTION,
                                ui.name, str(i==0), help=help))
        # i==0 to enable the first by default
    for ui in USER_INTERFACES.values():
        settings.extend(ui.default_settings())
    return settings

def load_ui(config, name=None):
    if name == None:
        uis = [c for c,v in config.items(USER_INTERFACE_SETTING_SECTION) if v == 'True']
        assert len(uis) == 1, 'Can only select one UI, not %d: %s' % (len(uis),uis)
        name = uis[0]
    ui = USER_INTERFACES[name]
    try:
        ui.config = dict(config.items(ui.setting_section))
    except configparser.NoSectionError:
        pass
    return ui
